/*
 * Project: VizKit
 * Version: 3.3.5
 
 * Date: 20141018
 * File: TrackLyrics.cpp
 *
 */

/***************************************************************************

Copyright (c) 2004-2014 Heiko Wichmann (http://www.imagomat.de/vizkit)


This software is provided 'as-is', without any expressed or implied warranty. 
In no event will the authors be held liable for any damages
arising from the use of this software.

Permission is granted to anyone to use this software for any purpose,
including commercial applications, and to alter it and redistribute it
freely, subject to the following restrictions:

1. The origin of this software must not be misrepresented; 
   you must not claim that you wrote the original software. 
   If you use this software in a product, an acknowledgment 
   in the product documentation would be appreciated 
   but is not required.

2. Altered source versions must be plainly marked as such, 
   and must not be misrepresented as being the original software.

3. This notice may not be removed or altered from any source distribution.

 ***************************************************************************/

#include "TrackLyrics.h"
#include "VisualConfiguration.h"
#include "VisualErrorHandling.h"
#include "VisualActorGraphics.h"
#include "VisualImage.h"
#include "VisualStageBox.h"
#include "VisualDataStore.h"
#include "VisualPreferences.h"
#include "VisualStyledString.h"
#include "VisualAnimation.h"
#include "VisualTimeline.h"
#include "VisualAnimationQueue.h"
#include "VisualTimeline.h"


using namespace VizKit;


TrackLyrics::TrackLyrics() 
{
    trackLyricsAutoScrollView = NULL;
}


TrackLyrics::~TrackLyrics()
{
	clear();
}


void TrackLyrics::show(uint32 anElapsedAudioTime, uint32 aRemainingAudioTime) 
{
    if (this->trackLyricsAutoScrollView)
    {
        float aProgressValue = (float)anElapsedAudioTime / (float)(anElapsedAudioTime + aRemainingAudioTime);
        this->trackLyricsAutoScrollView->draw(aProgressValue);
    }
}


VisualStringStyle TrackLyrics::getTrackLyricsStringStyle()
{
	VisualStringStyle stringStyle;
	
	VisualPreferences::getValue(VisualPreferences::kTrackLyricsFont, stringStyle.fontNameStr);
	
	stringStyle.fontSize = (float)(VisualPreferences::getValue(VisualPreferences::kTrackLyricsFontSize));

	stringStyle.fontColor.r = VisualPreferences::getValue(VisualPreferences::kTrackLyricsFontColorRedFloat);
	stringStyle.fontColor.g = VisualPreferences::getValue(VisualPreferences::kTrackLyricsFontColorGreenFloat);
	stringStyle.fontColor.b = VisualPreferences::getValue(VisualPreferences::kTrackLyricsFontColorBlueFloat);
	stringStyle.fontColor.red = VisualPreferences::getValue(VisualPreferences::kTrackLyricsFontColorRedFloat);
	stringStyle.fontColor.green = VisualPreferences::getValue(VisualPreferences::kTrackLyricsFontColorGreenFloat);
	stringStyle.fontColor.blue = VisualPreferences::getValue(VisualPreferences::kTrackLyricsFontColorBlueFloat);
	
	stringStyle.horizontalAlignment = kCenterAligned;
	
	return stringStyle;
}


void TrackLyrics::setTrackLyricsString(VisualStyledString &aTrackLyricsString)
{
    if (this->trackLyricsAutoScrollView == NULL)
    {
        this->trackLyricsAutoScrollView = new VisualTextAutoScrollView;
    }
    this->trackLyricsAutoScrollView->setStyledString(aTrackLyricsString);
}


void TrackLyrics::refresh()
{
    if (this->trackLyricsAutoScrollView != NULL)
    {
        this->trackLyricsAutoScrollView->refreshLayout();
    }
}


void TrackLyrics::animateOpacity(double currentPosition, void* userData)
{
    TrackLyrics *aTrackLyrics = reinterpret_cast<TrackLyrics*>(userData);
    VisualStringStyle aStringStyle = aTrackLyrics->getTrackLyricsStringStyle();
    RGBAColor aTextColor = aStringStyle.fontColor;
    aTextColor.r = aTextColor.red * currentPosition;
    aTextColor.g = aTextColor.green * currentPosition;
    aTextColor.b = aTextColor.blue * currentPosition;
    aTextColor.a = aTextColor.alpha * currentPosition;
    aStringStyle.fontColor = aTextColor;
    aTrackLyrics->trackLyricsAutoScrollView->setStringStyle(aStringStyle);
}


void TrackLyrics::fadeIn(uint32 durationInMilliseconds, bool startWithCurrentOpacityValue)
{
    double aStartOpacityValue = 0.0;
    if (startWithCurrentOpacityValue)
    {
        VisualStringStyle aStringStyle = this->trackLyricsAutoScrollView->getStringStyle();
        RGBAColor aFontColor = aStringStyle.fontColor;
        aStartOpacityValue = aFontColor.a;
    }
    VisualAnimation aFadeInAnimation(kCustomAnimatedProperty);
    aFadeInAnimation.setDurationInMilliseconds(durationInMilliseconds);
	aFadeInAnimation.setStartValue(aStartOpacityValue);
	aFadeInAnimation.setStopValue(1.0);
    aFadeInAnimation.setInterpolationType(kSinusoidalInterpolation);
	aFadeInAnimation.setAnimateCallbackFunctionPtr(TrackLyrics::animateOpacity, this);
	VisualAnimationQueue::removeVisualAnimationsWithOwnerIdentifier(this->opacityAnimationIdentifier, kCustomAnimatedProperty);
	VisualAnimationQueue::push(aFadeInAnimation, this->opacityAnimationIdentifier);
}


void TrackLyrics::fadeOut(uint32 durationInMilliseconds, double stopValue)
{
    VisualStringStyle aStringStyle = this->trackLyricsAutoScrollView->getStringStyle();
    RGBAColor aFontColor = aStringStyle.fontColor;
    double aCurrentOpacityValue = aFontColor.a;
    VisualAnimation aFadeOutAnimation(kCustomAnimatedProperty);
    aFadeOutAnimation.setDurationInMilliseconds(durationInMilliseconds);
	aFadeOutAnimation.setStartValue(aCurrentOpacityValue);
	aFadeOutAnimation.setStopValue(stopValue);
    aFadeOutAnimation.setInterpolationType(kSinusoidalInterpolation);
	aFadeOutAnimation.setAnimateCallbackFunctionPtr(TrackLyrics::animateOpacity, this);
	VisualAnimationQueue::removeVisualAnimationsWithOwnerIdentifier(this->opacityAnimationIdentifier, kCustomAnimatedProperty);
	VisualAnimationQueue::push(aFadeOutAnimation, this->opacityAnimationIdentifier);
}


void TrackLyrics::clear()
{
    VisualAnimationQueue::removeVisualAnimationsWithOwnerIdentifier(this->opacityAnimationIdentifier, kCustomAnimatedProperty);
    if (this->trackLyricsAutoScrollView)
    {
        delete this->trackLyricsAutoScrollView;
        this->trackLyricsAutoScrollView = NULL;
    }
}
