/*
 * Project: VizKit
 * Version: 3.3.5
 
 * Date: 20141018
 * File: VisualCustomization.h
 *
 */

/***************************************************************************

Copyright (c) 2004-2014 Heiko Wichmann (http://www.imagomat.de/vizkit)


This software is provided 'as-is', without any expressed or implied warranty. 
In no event will the authors be held liable for any damages
arising from the use of this software.

Permission is granted to anyone to use this software for any purpose,
including commercial applications, and to alter it and redistribute it
freely, subject to the following restrictions:

1. The origin of this software must not be misrepresented; 
   you must not claim that you wrote the original software. 
   If you use this software in a product, an acknowledgment 
   in the product documentation would be appreciated 
   but is not required.

2. Altered source versions must be plainly marked as such, 
   and must not be misrepresented as being the original software.

3. This notice may not be removed or altered from any source distribution.

 ***************************************************************************/

#ifndef VisualCustomization_h
#define VisualCustomization_h


/**
 * Customization values for individual visualizer plug-in development.
 */


/* The name of the visualizer plug-in, the identifier */
#define PluginIdentifierName "VizKit"

/**
 * The name of the visualizer as shown in menu of iTunes.
 * The name is set here with an UFT-8 encoded string. 
 * UTF-8 characters are set with escaped hex value.
 * Note that iTunes (especially on Windows) is not able to display all Unicode letters in the Menu.
 */
#define ShowName "VizKit"
// Test with KATAKANA LETTER ZU (Unicode: 30BA, UTF8: E3 82 BA)
//#define ShowName "Viz\xE3\x82\xBAKit"
// Test with GREEK CAPITAL LETTER OMEGA (Unicode: 03A9, UTF8: CE A9)
//#define ShowName "Viz\xCE\xA9Kit"
// Test with LATIN CAPITAL LETTER A WITH DIAERESIS (Unicode: 00C4, UTF8: C3 84)
//#define ShowName "Viz\xC3\x84Kit"

/**
 * The identifier of the visualizer plug-in (as used on Mac OS X for preferences identification and as CFBundleIdentifier).
 * Preferences can be stored with the Options Dialog. 
 * The values are stored in preferences text file. 
 * The preferences file is located at application data (Windows);
 * and on Mac OS X the preferences are stored in Preferences folder inside Library folder in user's home directory.
 * The domain identifier specifies the name of the plist file on Mac OS X.
 * On Windows the domain identifier specifies the name of the folder inside the application data directory.
 */
#if TARGET_OS_MAC
#define PluginDomainIdentifier "de.imagomat.vizkit"
#endif
#if TARGET_OS_WIN
#define PluginDomainIdentifier "Imagomat"
#endif


#if TARGET_OS_MAC
// graphic class identifiers need to be customized per plug-in on the Mac (DeImagomatVizKit)

// preventing name conflicts
// https://developer.apple.com/library/mac/#documentation/Cocoa/Conceptual/LoadingCode/Tasks/NameConflicts.html

#define VisualView DeImagomatVizKitVisualView
#define OptionsDialogWindowController DeImagomatVizKitOptionsDialogWindowController
#endif


/**
 * Visualizer plug-ins can display a configuration window.
 * Set this define to 1 if the visualizer displays a configuration window.
 * Set this define to 0 if the visualizer does not display a configuration window.
 */
#define HasConfigurationWindow 1


/** The major release number of the visualizer (version number). Range: 0 - 99. For instance 1.2.5 means: major release number 1, minor release number 2, revision number 5. */
#define MajorReleaseNumber 3

/** The minor release number of the visualizer (version number). Range: 0 - 9. For instance 1.2.5 means: major release number 1, minor release number 2, revision number 5. */
#define MinorReleaseNumber 3

/** The revision number of the visualizer (version number). Range: 0 - 255. For instance 1.2.5 means: major release number 1, minor release number 2, revision number 5. */
#define ReleaseRevisionNumber 5

/**
 * The URL of the current version information for this visualizer.
 * If set to an empty string, no check for update information is performed.
 *
 * The remote text information contains three items, separated by "|": 
 * 1) The current version number (major.minor.revision), set each component of the version (including 0). Zero values at the end of the version are dropped when presenting the current version to the user.
 * 2) Info URL (web page with general information about the visualizer or specific information about the advertized version)
 * 3) Download URL (URL pointing to download file of current visualizer)
 * This information is used to prompt the user with update information in case an older version is running.
 * The feature is turned on/off by preference VisualPreferences::kCheckForUpdate.
 */
#define CurrentVersionInformationURL "http://www.imagomat.de/vizkit/download/current_version.txt"


/**
 * Additional keys for notifications.
 * The keys are added to the enum VisualNotificationKey in VisualNotificationKey.h
 * The notifications are dispatched and sent to each actor.
 */
#define AdditionalDispatchedNotificationKeys

/**
 * Code block with case switches for the added notification keys.
 * The code block is inserted in the switch case in VisualMainAction::renderAction()
 */
#define AdditionalDispatchedNotificationCasesCode

/**
 * Code block with case switches for the added notification keys.
 * The code block is inserted in the switch case in VisualNotification::convertNotificationKeyToString()
 */
#define AdditionalDispatchedNotificationToStringCasesCode


/**
 * Additional keys for preferences.
 * The keys are added to the enums in VisualPreferences.h
 */
#define AdditionalPreferenceKeyIntKeys

#define AdditionalPreferenceKeyFloatKeys

#define AdditionalPreferenceKeyCharKeys

#define AdditionalPreferenceKeyBoolKeys

/**
 * Code block with case switches for the added preference keys.
 * The code blocks are added to the convert methods (string to key and key to string)
 */

#define AdditionalPreferenceKeyIntToKeyCasesCode

#define AdditionalPreferenceKeyFloatToKeyCasesCode

#define AdditionalPreferenceKeyCharToKeyCasesCode

#define AdditionalPreferenceKeyBoolToKeyCasesCode

#define AdditionalPreferenceKeyIntToStringCasesCode

#define AdditionalPreferenceKeyFloatToStringCasesCode

#define AdditionalPreferenceKeyCharToStringCasesCode

#define AdditionalPreferenceKeyBoolToStringCasesCode


namespace VizKit {
	
	/**
	 * Collection of individual customizable functions.
     * Each visualizer plug-in can implement its own version of these functions.
	 * All functions of VisualCustomization are static.
     * VisualCustomization needs no state.
	 */
	class VisualCustomization {
        
	public:
        
        /**
         * Initializes the actors and adds them to the ensemble.
         * Called by VisualStageControl on init.
         */
        static void initEnsembleMembers(void);
        
        /**
         * Override and add default preference values.
         */
        static void setDefaultPreferenceValues(void);
        
        /**
         * Processes keyboard events.
         * @param aKeyboadCharacterValue The value of the key that was pressed.
         * @return True if key press event was handled by visualizer, false if key press event was not handled by visualizer.
         * @remarks This method is only called when the visualizer is showing.
         */
        static bool handleKeyPressed(const char aKeyboadCharacterValue);

	private:
        
		/**
		 * The constructor.
		 */
		VisualCustomization();
		
		/**
		 * The destructor.
		 */
		~VisualCustomization();
		
	};
    
}



#endif /* VisualCustomization_h */

