/*
 * Project: VizKit
 * Version: 3.3.5
 
 * Date: 20141018
 * File: VisualMainAction.h
 *
 */

/***************************************************************************
 
 Copyright (c) 2004-2014 Heiko Wichmann (http://www.imagomat.de/vizkit)
 
 
 This software is provided 'as-is', without any expressed or implied warranty. 
 In no event will the authors be held liable for any damages
 arising from the use of this software.
 
 Permission is granted to anyone to use this software for any purpose,
 including commercial applications, and to alter it and redistribute it
 freely, subject to the following restrictions:
 
 1. The origin of this software must not be misrepresented; 
 you must not claim that you wrote the original software. 
 If you use this software in a product, an acknowledgment 
 in the product documentation would be appreciated 
 but is not required.
 
 2. Altered source versions must be plainly marked as such, 
 and must not be misrepresented as being the original software.
 
 3. This notice may not be removed or altered from any source distribution.
 
 ***************************************************************************/

#include "VisualMainAction.h"

#include <time.h> // for srand()

#include "VisualConfiguration.h"
#include "VisualConfigurationDialog.h"
#include "VisualErrorHandling.h"
#include "VisualPreferences.h"
#include "VisualDataStore.h"
#include "VisualGraphics.h"
#include "VisualAudioLab.h"
#include "VisualTiming.h"
#include "VisualStageControl.h"
#include "VisualAnimationQueue.h"
#include "VisualNotification.h"
#include "VisualNotificationQueue.h"
#include "VisualThreading.h"
#include "VisualThreadingManager.h"
#include "VisualStyledString.h"
#include "VisualImage.h"
#include "VisualFontManager.h"
#include "VisualFile.h"
#include "VisualNetwork.h"
#include "VisualUpdateManager.h"
#include "VisualActor.h"
#include "VisualSignature.h"
#include "VisualAudioData.h"
#include "VisualCustomization.h"


using namespace VizKit;


#if TARGET_OS_WIN

static	WNDPROC giTunesWndProc = NULL;

LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	LRESULT aResult = TRUE;
	bool doPassEvent = true;

	switch (message)
	{
		case WM_CHAR:
			// catch key pressed messages
			{
				bool didHandleKeyPress = VisualMainAction::handleKeyPressEvent((char)wParam);
				if (didHandleKeyPress)
				{
					doPassEvent = false;
				}
			}
                break;
	}
    
	if (doPassEvent)
	{
		aResult = CallWindowProc(giTunesWndProc, hWnd, message, wParam, lParam);
	}

	return aResult;
}


#endif


void VisualMainAction::updateTiming() {
	VisualTiming::update();
}


void VisualMainAction::setAppVersion(int majorRev, int minorAndBugRev) 
{
    int minorRev = minorAndBugRev >> 4;
    int bugRev = minorAndBugRev & 0x0F;
	VisualDataStore::setAppVersion(majorRev, minorRev, bugRev);
    
    if ((VisualDataStore::getAppVersionMajorNum() < 10) || ((VisualDataStore::getAppVersionMajorNum() == 10) && (VisualDataStore::getAppVersionMinorNum() < 4)))
    {
        char errLog[512];
        sprintf(errLog, "Err: Visualizer Plugin %s needs iTunes 10.4 or higher to run", VisualConfiguration::visualizerPluginIdentifierName());
        writeLog(errLog);
        VisualPlayerState* theVisualPlayerState = VisualPlayerState::getInstance();
		theVisualPlayerState->setPlayerShowMode(kErrorState);
    }
}


void VisualMainAction::setVisualizerAPIVersion(int majorVersionNum, int minorVersionNum) {
    VisualDataStore::setVisualizerAPIVersion(majorVersionNum, minorVersionNum);
}


void VisualMainAction::handleInitMessage(int maxDataChannels, int numberOfWaveformEntries, int numberOfSpectrumEntries) {
	
	VisualAudioLab::setMaxNumberOfAudioDataChannels(maxDataChannels);
	VisualAudioLab::setNumberOfAudioWaveformDataEntries(numberOfWaveformEntries);
	VisualAudioLab::setNumberOfAudioSpectrumDataEntries(numberOfSpectrumEntries);

	VisualPreferences::setDefaultValues();
	
}


void VisualMainAction::setPlatformView(VISUAL_PLATFORM_VIEW aView) {
	VisualGraphics* theVisualGraphics = VisualGraphics::getInstance();
	theVisualGraphics->setPlatformView(aView);
}


void VisualMainAction::handleShowWindowMessage(bool isFullScreen) {

	VisualPlayerState* theVisualPlayerState = VisualPlayerState::getInstance();
    if (theVisualPlayerState->getPlayerShowMode() == kErrorState) {
        return;
    }
	if (isFullScreen) {
		theVisualPlayerState->setPlayerShowMode(kIsShowingInFullScreen);
    } else {
		theVisualPlayerState->setPlayerShowMode(kIsShowingInWindow);
    }
    
	VisualGraphics* theVisualGraphics = VisualGraphics::getInstance();
	theVisualGraphics->isSetupForFullScreenMode(isFullScreen);

	VisualMainAction::handleShowShouldStartMessage(isFullScreen);

	if (theVisualPlayerState->getPlayerShowMode() & kIsShowing) 
	{
#if TARGET_OS_WIN
		if (giTunesWndProc == NULL)
		{
			VisualGraphics* theVisualGraphics = VisualGraphics::getInstance();
			VISUAL_PLATFORM_VIEW aPlatformView = theVisualGraphics->getPlatformView();
			giTunesWndProc = (WNDPROC) SetWindowLong(aPlatformView, GWL_WNDPROC, (LONG)WndProc);
		}
		VisualMainAction::renderAction();
#endif
	}
}


#if TARGET_OS_MAC
void VisualMainAction::startShow () {
    VisualGraphics* theVisualGraphics = VisualGraphics::getInstance();
    theVisualGraphics->startRendering();
}
#endif


void VisualMainAction::handleSetWindowMessage(bool isFullScreen) {
    VisualGraphics* theVisualGraphics = VisualGraphics::getInstance();
	theVisualGraphics->isSetupForFullScreenMode(isFullScreen);
	
	VisualMainAction::handleCanvasReshape();
	
	VisualPlayerState* theVisualPlayerState = VisualPlayerState::getInstance();
	if (theVisualPlayerState->getPlayerShowMode() & kIsShowing) {
#if TARGET_OS_WIN
		VisualMainAction::renderAction();
#endif
	}
}


void VisualMainAction::handlePulseMessage(int currentPositionInMS, int timeStampID, int numberOfWaveformChannels, int numberOfWaveformEntries, char* waveformData, int numberOfSpectrumChannels, int numberOfSpectrumEntries, char* spectrumData) {
	
	static bool isProcessingAudioData = false;
	
	VisualPlayerState* theVisualPlayerState = VisualPlayerState::getInstance();
    if (theVisualPlayerState->getPlayerShowMode() == kErrorState) 
	{
        return;
    }
	
	if (!isProcessingAudioData)
	{
		isProcessingAudioData = true;
		
		VisualMainAction::setTrackPlayPositionInMS(currentPositionInMS);
		
		VisualAudioLab* theVisualAudioLab = VisualAudioLab::getInstance();
		if (timeStampID != (int)(theVisualAudioLab->getCurrTimestampIDOfAudioData())) 
		{
			if ((numberOfWaveformChannels > 0) && (numberOfSpectrumChannels > 0) && (numberOfWaveformEntries > 0) && (numberOfSpectrumEntries > 0))
			{                
                VisualAudioData *visualAudioData = new VisualAudioData;
                visualAudioData->setNumberOfWaveformChannels(numberOfWaveformChannels);
                visualAudioData->setNumberOfWaveformEntries(numberOfWaveformEntries);
                visualAudioData->setWaveformData((const uint8* const)waveformData);
                visualAudioData->setNumberOfSpectrumChannels(numberOfSpectrumChannels);
                visualAudioData->setNumberOfSpectrumEntries(numberOfSpectrumEntries);
                visualAudioData->setSpectrumData((const uint8* const)spectrumData);
                
                VisualNotification anAudioDataMessageNotification;
                anAudioDataMessageNotification.setPointer(visualAudioData);
                anAudioDataMessageNotification.setKey(kAudioDataMsg);
                anAudioDataMessageNotification.post();
				
				theVisualAudioLab->setCurrTimestampIDOfAudioData(timeStampID);
			}
		}
		isProcessingAudioData = false;
	}

#if TARGET_OS_WIN
	if (theVisualPlayerState->getPlayerShowMode() & kIsShowing) 
	{
		VisualMainAction::renderAction();
	}
#endif

}


void VisualMainAction::resetVisualAudioLabDataIfNotPaused() {
	VisualPlayerState* theVisualPlayerState = VisualPlayerState::getInstance();
	if ((theVisualPlayerState->getAudioPlayState() & kAudioPlayPaused) != kAudioPlayPaused) {
		VisualAudioLab* theVisualAudioLab = VisualAudioLab::getInstance();
		theVisualAudioLab->resetData();
	}
}


bool VisualMainAction::prepareRenderAction() {

	VisualGraphics* theVisualGraphics = VisualGraphics::getInstance();

#if TARGET_OS_WIN
	theVisualGraphics->initGdiplus();
#endif

	bool success = theVisualGraphics->setupContext();
	if (!success) {
		return success;
	}
	
	theVisualGraphics->setCurrentContext();
	
	theVisualGraphics->setCanvasViewport();
	theVisualGraphics->clearCanvasBackground();

	theVisualGraphics->gatherOpenGLCapabilities();

#if TARGET_OS_WIN
	theVisualGraphics->buildFont();
#endif

	theVisualGraphics->resetModelViewMatrix();
	
	VisualMainAction::postTrackInfoOfCurrentAudioData();
	VisualMainAction::requestAlbumCoverArtworkOfCurrentAudioData();
	success = VisualMainAction::postLyricsOfCurrentAudioData();
	
	return success;

}


void VisualMainAction::renderAction() {

	static bool isRendering = false;
	if (isRendering == true)
	{
        char errLog[512];
        sprintf(errLog, "Warn: Dropping frame in file: %s (line: %d) [%s])", __FILE__, __LINE__, __FUNCTION__);
        writeLog(errLog);
		return;
	}
	isRendering = true;
	
	bool debug = false;
	
	VisualThreadingManager::update();

	VisualPlayerState* theVisualPlayerState = VisualPlayerState::getInstance();
	VisualGraphics* theVisualGraphics = VisualGraphics::getInstance();

	theVisualGraphics->setCurrentContext();
	theVisualGraphics->clearCanvasBackground();

	VisualStageControl::checkForFadeOutEvent();
	
	VisualNotificationQueue::update();
	
    VisualNetwork::startDownloadThreadIfNeeded();
	
	size_t notificationCount = 0;
	if (debug == true) {
		notificationCount = VisualNotificationQueue::size();
	}
	
	uint16 currNotificationIdx = 1;

	VisualNotification aNotification;
	VisualNotificationKey aKey;
	
	while (VisualNotificationQueue::size() > 0) {
	
		aNotification = VisualNotificationQueue::pop();
		aKey = aNotification.getKey();

		if (debug == true) {
			char notificationStr[64];
			char logStr[64];
			VisualNotification::convertNotificationKeyToString(aKey, notificationStr);
			sprintf(logStr, "Notification (%d/%ld): %s", currNotificationIdx, notificationCount, notificationStr);
			writeLog(logStr);
			currNotificationIdx++;
		}

		switch (aKey) {
			case kAudioPlayTrackChangedEvt:
			case kAudioPlayStartedEvt:
			case kAudioPlayStoppedEvt:
			case kAudioPlayPausedEvt:
			case kAudioPlayResumedEvt:
			case kAudioPlayReachedFadeOutTimeEvt:
			case kTrackPlayPositionChangedEvt:
			case kBeatImpulseEvt:
			case kTrackInfoImageMsg:
            case kTrackLyricsStringMsg:
			case kTrackLyricsImageMsg:
			case kImageHistogramCompletedMsg:
			case kTrackInfoTextureChangedMsg:
			case kCanvasReshapeEvt:
                
                AdditionalDispatchedNotificationCasesCode
                    
                VisualStageControl::dispatchNotification(aNotification);
                    
                if (aKey == kTrackLyricsStringMsg)
                {
                    VisualStyledString* styledTrackLyricsString = dynamic_cast<VisualStyledString*>(aNotification.getObject());
                    if (styledTrackLyricsString) {
                        VisualImage styledTrackLyricsStringImage;
                        bool success = styledTrackLyricsStringImage.initWithStyledString(*styledTrackLyricsString);
                        if (success) {
                            VisualNotification trackLyricsImageNotification;
                            trackLyricsImageNotification.setKey(kTrackLyricsImageMsg);
                            trackLyricsImageNotification.setObject(styledTrackLyricsStringImage);
                            trackLyricsImageNotification.post();
                        }
                    }
                }
                
                break;
                
            case kAudioDataMsg:
            {
                VisualAudioData* visualAudioData = reinterpret_cast<VisualAudioData*>(aNotification.getPointer());
                VisualAudioLab* theVisualAudioLab = VisualAudioLab::getInstance();
                theVisualAudioLab->processAudioData((const VisualAudioData&)*visualAudioData);
                theVisualAudioLab->checkForBeatImpulse();
                delete visualAudioData;
            }   
                break;
            
            case kImageWithIdentifierLoadedAndCreatedMsg:
            {
                VisualStageControl::dispatchNotification(aNotification);
                VisualImage* anImage = reinterpret_cast<VisualImage*>(aNotification.getPointer());
                if (anImage != NULL)
                {
                    delete anImage;
                }
				break;
            }
				
			case kLyricsAreWrittenIntoTempFileMsg:
			{
				const VisualFile* const lyricsFile = dynamic_cast<const VisualFile* const>(aNotification.getObject());
				if (lyricsFile) {
					VisualDataStore::fileWithLyricsStringId = lyricsFile->getIdentifier();
					const VisualFile* const lyricsFileCopy = lyricsFile->clone();
					bool success = VisualThreading::createThread((ThreadingFuncPtr)VisualDataStore::readFileAndRemoveFileAndSendNotificationWithString, (void*)lyricsFileCopy);
					if (!success)
					{
						char errLog[512];
						sprintf(errLog, "Err: Unable to create thread after kLyricsAreWrittenIntoTempFileMsg in file: %s (line: %d) [%s])", __FILE__, __LINE__, __FUNCTION__);
						writeLog(errLog);
					}
				}
			}
				break;
			case kStringWithIdentifierLoadedAndCreatedMsg:
				if (aNotification.getIdentifier() == VisualDataStore::fileWithLyricsStringId) {
					VisualItemIdentifier currentTrackIdentifier = VisualDataStore::getIdentifierOfCurrentTrack();
					if (currentTrackIdentifier == VisualDataStore::trackIdentifierOfLyricsMetadata) {
						// audio track is not different from the one we initiated the lyrics retrieval for
						const VisualString* const lyricsString = dynamic_cast<const VisualString* const>(aNotification.getObject());
						if (lyricsString) {
							VisualDataStore::setLyricsOfCurrentTrack(*lyricsString);
							VisualNotification::post(kLyricsAreAvailableInMetadataMsg);
						}
					}
				} else if (aNotification.getIdentifier() == VisualDataStore::updateInformationStringId) {
					bool updateIsAvailable = false;
					updateIsAvailable = VisualUpdateManager::evaluateUpdateInformation();
					if (updateIsAvailable) {
						VisualActor* updateServiceActor = VisualStageControl::getVisualActorByName("UPDATESERVICE");
						if (updateServiceActor) {
							updateServiceActor->setState(kVisActOn);
						}
					}
				} else {
					VisualStageControl::dispatchNotification(aNotification);
				}
				break;
			case kLyricsAreAvailableInMetadataMsg:
			{
                VisualString trackLyrics;
				const VisualString aCurrentTrackLyricsString = VisualDataStore::getLyricsOfCurrentTrack();
				// "missing value" is literally sent with streams that do not convey any lyrics
				const VisualString missingValueString("missing value");
				const VisualString missingValueStringWithLineEndings("missing value\n");
                if ((aCurrentTrackLyricsString != missingValueString) || (aCurrentTrackLyricsString != missingValueStringWithLineEndings))
                {
                    trackLyrics = aCurrentTrackLyricsString;
                }
                VisualNotification trackLyricsStringStyleNotification;
                trackLyricsStringStyleNotification.setKey(kGetTrackLyricsStringStyleMsg);
                VisualStageControl::dispatchNotification(trackLyricsStringStyleNotification);
                VisualStyledString styledTrackLyricsString;
                styledTrackLyricsString.initWithVisualString(trackLyrics);
                const VisualStringStyle* const trackLyricsStringStyle = dynamic_cast<const VisualStringStyle* const>(trackLyricsStringStyleNotification.getObject());
                if (trackLyricsStringStyle) {
                    styledTrackLyricsString.setStyle(*trackLyricsStringStyle);
                    
                    VisualNotification styledTrackLyricsStringNotification;
                    styledTrackLyricsStringNotification.setKey(kTrackLyricsStringMsg);
                    styledTrackLyricsStringNotification.setObject(styledTrackLyricsString);
                    styledTrackLyricsStringNotification.post();
                }
			}
				break;
			case kStyledStringMsg:
				if (aNotification.getIdentifier() == VisualDataStore::styledTrackInfoStringId) {
					VisualStyledString* styledTrackInfoString = dynamic_cast<VisualStyledString*>(aNotification.getObject());
					if (styledTrackInfoString) {
						VisualImage styledTrackInfoStringImage;
						bool success = styledTrackInfoStringImage.initWithStyledString(*styledTrackInfoString);
						if (success) {
							VisualNotification trackInfoImageNotification;
							trackInfoImageNotification.setKey(kTrackInfoImageMsg);
							trackInfoImageNotification.setObject(styledTrackInfoStringImage);
							trackInfoImageNotification.post();
						}
					}
                }
				break;
			case kLoadingEncodedImageDataCompletedMsg:
			{
				VisualImage* visualImage = reinterpret_cast<VisualImage*>(aNotification.getPointer());
				if (visualImage) {
					bool success = visualImage->initWithLoadedEncodedData();
					if (success) {
						VisualNotification imageCreationCompletedNotification(aNotification.getIdentifier());
						imageCreationCompletedNotification.setKey(kImageWithIdentifierLoadedAndCreatedMsg);
                        imageCreationCompletedNotification.setPointer(visualImage);
						imageCreationCompletedNotification.post();
					}
				}
			}
				break;
			case kImageWriteToPNGFileMsg:
			case kImageWriteToPNGFileAndDeleteMsg:
			{
				VisualImage* visualImage = reinterpret_cast<VisualImage*>(aNotification.getPointer());
				if (visualImage) {
					bool success = false;
					VisualFile* aFile = dynamic_cast<VisualFile*>(aNotification.getObject());
					success = visualImage->writeToPNGFile(*aFile);
                    if (success == false)
                    {
                        char errStr[512];
                        sprintf(errStr, "Err: Failure on writeToPNGFile in file: %s (line: %d) [%s])", __FILE__, __LINE__, __FUNCTION__);
                        writeLog(errStr);
                    }
					if (aKey == kImageWriteToPNGFileAndDeleteMsg) {
						delete visualImage;
					}
				}
			}
				break;
			default:
				char notificationString[32];
				VisualNotification::convertNotificationKeyToString(aKey, notificationString);
				char errLog[512];
				sprintf(errLog, "Unhandled VisualNotificationKey %s in file: %s (line: %d) [%s])", notificationString, __FILE__, __LINE__, __FUNCTION__);
				writeLog(errLog);
		}

	}

	AudioPlayState playState = theVisualPlayerState->getAudioPlayState();
	if (playState == kAudioPlayStarted) {
		uint32 fadeInTimeOnPlayInMS = VisualPreferences::getValue(VisualPreferences::kFadeInTimeOnPlayInMS);
		if (theVisualPlayerState->getElapsedAudioPlayStartTime() > fadeInTimeOnPlayInMS) {
			theVisualPlayerState->setAudioPlayState(kAudioIsPlaying);
		}
	} else if (playState == kAudioPlayResumed) {
		uint32 fadeInTimeOnResumeInMS = VisualPreferences::getValue(VisualPreferences::kFadeInTimeOnResumeInMS);
		if (theVisualPlayerState->getElapsedAudioPlayStartTime() > fadeInTimeOnResumeInMS) {
			theVisualPlayerState->setAudioPlayState(kAudioIsPlaying);
		}
	}

    VisualMainAction::monitorRenderMessageProcess();

	// show action
	VisualStageControl::doEnsembleShow(*theVisualPlayerState);
	
	theVisualGraphics->finishGLDrawing();
    
#if TARGET_OS_MAC	
	if ((theVisualPlayerState->getPlayerShowMode() & kIsShowingInFullScreen) == kIsShowingInFullScreen) {
		OSErr err = noErr;
		err = UpdateSystemActivity(UsrActivity); // indicate user activity to prevent screen saver or sleep mode from kicking in
        if (err != noErr)
        {
            char errStr[512];
            sprintf(errStr, "Err: Failure on UpdateSystemActivity in file: %s (line: %d) [%s])", __FILE__, __LINE__, __FUNCTION__);
            writeLog(errStr);
        }
	}
#endif
	
	isRendering = false;
	
}


void VisualMainAction::finishRenderAction() {
    VisualGraphics* theVisualGraphics = VisualGraphics::getInstance();
    theVisualGraphics->disposeContext();
	VisualStageControl::dispose();
	VisualAnimationQueue::dispose();
	VisualNotificationQueue::dispose();
#if TARGET_OS_WIN
	theVisualGraphics->terminateGdiplus();
#endif
	VisualGraphics::dispose();
}


void VisualMainAction::monitorRenderMessageProcess() {
	
	static float framesPerSecond = 0.0f;
	static uint16 frameCount = 0;
	uint32 elapsedMilliseconds = 0;
	uint32 remainingMilliseconds = 0;
	VisualPlayerState* theVisualPlayerState = NULL;
	VisualGraphics* theVisualGraphics = NULL;
	uint8 playState;
	char cStr64[64];
	
	theVisualPlayerState = VisualPlayerState::getInstance();
	
    elapsedMilliseconds = VisualTiming::getElapsedMilliSecsSinceReset("fps");
    
    if (elapsedMilliseconds > 1000) {
        framesPerSecond = (float)frameCount / (float)elapsedMilliseconds * 1000.0f;
        frameCount = 0;
        VisualTiming::resetTimestamp("fps");
    }
	frameCount++;
    
    sprintf(cStr64, "%.2f",  framesPerSecond);
    VisualDataStore::setProcessInfo("FramesPerSecond", cStr64);
	
	theVisualGraphics = VisualGraphics::getInstance();
	
	VisualDataStore::setProcessInfo("RendererName",theVisualGraphics->getRendererName());
	
	VisualDataStore::setProcessInfo("RendererVendor",theVisualGraphics->getRendererVendor());
	
	VisualDataStore::setProcessInfo("RendererVersion",theVisualGraphics->getRendererVersion());
	
	sprintf(cStr64, "%#x", theVisualGraphics->getGLVersion());
	VisualDataStore::setProcessInfo("GLVersion", cStr64);
	
	sprintf(cStr64, "%ld", theVisualGraphics->getMaxTextureSize());
	VisualDataStore::setProcessInfo("Maximum Texture Size", cStr64);
	
	// screen measurements
	PixelRect canvasRect = theVisualGraphics->getCanvasRect();
	sprintf(cStr64, "width: %lu, height: %lu", (unsigned long)canvasRect.width, (unsigned long)canvasRect.height);
	VisualDataStore::setProcessInfo("Canvas", cStr64);
	
	/*
	 sprintf(cStr64, "top: %.2f, left: %.2f", theVisualGraphics->getMaxTopCoordOfCanvas(), theVisualGraphics->getMaxLeftCoordOfCanvas());
	 VisualDataStore::setProcessInfo("CoordMaxTopLeft", cStr64);
	 sprintf(cStr64, "bottom: %.2f, right: %.2f", theVisualGraphics->getMaxBottomCoordOfCanvas(), theVisualGraphics->getMaxRightCoordOfCanvas());
	 VisualDataStore::setProcessInfo("CoordMaxBottomRight", cStr64);
	 sprintf(cStr64, "near: %.2f, far: %.2f", theVisualGraphics->getMaxNearCoordOfCanvas(), theVisualGraphics->getMaxFarCoordOfCanvas());
	 VisualDataStore::setProcessInfo("CoordMaxNearFar", cStr64);
	 */
	
	BottomLeftPositionedPixelRect bottomLeftPositionedPixelRect = theVisualGraphics->getViewportRect();
	sprintf(cStr64, "bottom: %lu, left: %lu, width: %lu, height: %lu",  bottomLeftPositionedPixelRect.bottomLeftPixel.y, bottomLeftPositionedPixelRect.bottomLeftPixel.x, bottomLeftPositionedPixelRect.pixelRect.width, bottomLeftPositionedPixelRect.pixelRect.height);
	VisualDataStore::setProcessInfo("ViewportRect", cStr64);
	
	playState = theVisualPlayerState->getAudioPlayState();
	switch (playState) {
		case kAudioIsPlaying:
			sprintf(cStr64, "kAudioIsPlaying");
			break;
		case kAudioPlayStarted:
			sprintf(cStr64, "kAudioPlayStarted");
			break;		
		case kAudioPlayResumed:
			sprintf(cStr64, "kAudioPlayResumed");
			break;
		case kAudioPlayPaused:
			sprintf(cStr64, "kAudioPlayPaused");
			break;
		case kAudioIsNotPlaying:
			sprintf(cStr64, "kAudioIsNotPlaying");
			break;
		default:
			sprintf(cStr64, "unknown");
	}
	VisualDataStore::setProcessInfo("AudioPlayState", cStr64);
	
	elapsedMilliseconds = theVisualPlayerState->getElapsedAudioTime();
	sprintf(cStr64, "%d", elapsedMilliseconds);
	VisualDataStore::setProcessInfo("TrackTimeElapsed", cStr64);
	
	remainingMilliseconds = theVisualPlayerState->getRemainingAudioTime();
	sprintf(cStr64, "%d", remainingMilliseconds);
	VisualDataStore::setProcessInfo("TrackTimeRemaining", cStr64);
	
}


void VisualMainAction::handleHideWindowMessage() 
{
#if TARGET_OS_WIN
	if (giTunesWndProc != NULL)
	{
		VisualGraphics* theVisualGraphics = VisualGraphics::getInstance();
		VISUAL_PLATFORM_VIEW aPlatformView = theVisualGraphics->getPlatformView();
		SetWindowLong(aPlatformView, GWL_WNDPROC, (LONG)giTunesWndProc);	
		giTunesWndProc = NULL;
	}
#endif

	VisualPlayerState* theVisualPlayerState = VisualPlayerState::getInstance();
    theVisualPlayerState->setPlayerShowMode(kIsNotShowing);
	
	VisualMainAction::finishRenderAction();
}


void VisualMainAction::handleAudioPlayStoppedEvent() {

	VisualPlayerState* theVisualPlayerState = VisualPlayerState::getInstance();
	
	VisualTiming::resetTimestamp("trackPlayStop");
	
	if (theVisualPlayerState->remainingAudioTimeIsKnown() == true && theVisualPlayerState->getRemainingAudioTime() > 2000) {
		// if more than 2 secs. of audio is left to be played,
		// we assume that playing paused
		theVisualPlayerState->setAudioPlayState(kAudioPlayPaused);
		if (theVisualPlayerState->getPlayerShowMode() & kIsShowing) {
			VisualNotification::post(kAudioPlayPausedEvt);
		}
	} else {
		theVisualPlayerState->setAudioPlayState(kAudioIsNotPlaying);
		VisualMainAction::setTrackPlayPositionInMS(0);
		if (theVisualPlayerState->getPlayerShowMode() & kIsShowing) {
			VisualNotification::post(kAudioPlayStoppedEvt);
		}
	}
}


void VisualMainAction::handleAudioPlayStartedEvent(bool trackDidChange) {

	bool success = false;

	VisualPlayerState* theVisualPlayerState = VisualPlayerState::getInstance();
	AudioPlayState prevAudioPlayState = theVisualPlayerState->getAudioPlayState();
	
	VisualTiming::resetTimestamp("trackPlayStart");
	
	if (prevAudioPlayState == kAudioIsNotPlaying) {
		theVisualPlayerState->setAudioPlayState(kAudioPlayStarted);
	} else {
		if (trackDidChange == true) {
			VisualMainAction::setTrackPlayPositionInMS(0);
			theVisualPlayerState->setAudioPlayState(kAudioPlayStarted);
		} else {
			theVisualPlayerState->setAudioPlayState(kAudioPlayResumed);
		}
	}

	if (theVisualPlayerState->getPlayerShowMode() & kIsShowing) {
		if (theVisualPlayerState->getAudioPlayState() == kAudioPlayResumed) {
			VisualNotification::post(kAudioPlayResumedEvt);
		} else {
			if (trackDidChange) {
				VisualNotification::post(kAudioPlayTrackChangedEvt);
				VisualMainAction::postTrackInfoOfCurrentAudioData();
                VisualMainAction::requestAlbumCoverArtworkOfCurrentAudioData();
				success = VisualMainAction::postLyricsOfCurrentAudioData();
                if (success == false)
                {
                    char errStr[512];
                    sprintf(errStr, "Err: Failure on postLyricsOfCurrentAudioData in file: %s (line: %d) [%s])", __FILE__, __LINE__, __FUNCTION__);
                    writeLog(errStr);
                }
			} else {
				VisualNotification::post(kAudioPlayStartedEvt);
			}
		}
	}

}


void VisualMainAction::setTrackPlayPositionInMS(int trackPlayPosition) {
	static VisualItemIdentifier currentTrackIdentifier;
	static int prevTrackPlayPosition = 0;
	
	VisualPlayerState* theVisualPlayerState = VisualPlayerState::getInstance();
	theVisualPlayerState->setTrackPlayPositionInMS(trackPlayPosition);

	if (currentTrackIdentifier == VisualDataStore::getIdentifierOfCurrentTrack()) {
		if ( ((trackPlayPosition - prevTrackPlayPosition) * (trackPlayPosition - prevTrackPlayPosition)) > (5000 * 5000)) {
			// no track play position change since 5000 MS -> position changed by random access (user changed play position manually)
			VisualNotification::post(kTrackPlayPositionChangedEvt);
		}
	}
	currentTrackIdentifier = VisualDataStore::getIdentifierOfCurrentTrack();
	prevTrackPlayPosition = trackPlayPosition;
}


void VisualMainAction::handleShowShouldStartMessage(bool isFullScreen) {
    
    srand((unsigned int)time(NULL)); // Seed random number generator

	bool success = VisualMainAction::prepareRenderAction();

	if (!success) {
		VisualPlayerState* theVisualPlayerState = VisualPlayerState::getInstance();
		theVisualPlayerState->setPlayerShowMode(kErrorState);
	} else {
		VisualUpdateManager::checkForUpdate();
	}
	
}


void VisualMainAction::closeThreads() {
	VisualNetwork::dispose();
	VisualThreadingManager::dispose();
	VisualUpdateManager::dispose();
}


void VisualMainAction::closePersistentStore() {
	VisualDataStore::dispose();
	VisualConfiguration::dispose();
	VisualPreferences::dispose();
	VisualPlayerState::dispose();
	VisualConfigurationDialog::dispose();
	VisualFontManager::dispose();
	VisualTiming::dispose();
	VisualAudioLab::dispose();
}


void VisualMainAction::handleCanvasReshape() {
	VisualGraphics* theVisualGraphics = VisualGraphics::getInstance();
    theVisualGraphics->refreshCanvasRect();
	theVisualGraphics->setCurrentContext();
	theVisualGraphics->setCanvasViewport();
	theVisualGraphics->clearCanvasBackground();
	VisualNotification::post(kCanvasReshapeEvt);
}


bool VisualMainAction::handleKeyPressEvent(char aKeyPressed) 
{
    bool didHandleTheKeyPress = false;
	VisualPlayerState* theVisualPlayerState = VisualPlayerState::getInstance();
	if (theVisualPlayerState->getPlayerShowMode() & kIsShowing) 
	{
        didHandleTheKeyPress = VisualCustomization::handleKeyPressed(aKeyPressed);
	}
    return didHandleTheKeyPress;
}


void VisualMainAction::showConfigurationDialog() {
	VisualConfigurationDialog::show();
}


void VisualMainAction::resetCurrAudioMetaData(void) {
	VisualDataStore::resetCurrAudioMetaData();
}


void VisualMainAction::setAudioTrackName(const void* const audioTrackName, int audioTrackNameLength) {
	VisualDataStore::setAudioTrackName((const uint16*)audioTrackName, audioTrackNameLength);
}


void VisualMainAction::setAudioTrackArtistName(const void* const audioTrackArtistName, int audioTrackArtistNameLength) {
	VisualDataStore::setAudioTrackArtistName((const uint16*)audioTrackArtistName, audioTrackArtistNameLength);
}


void VisualMainAction::setAudioTrackAlbumName(const void* const audioTrackAlbumName, int audioTrackAlbumNameLength) {
	VisualDataStore::setAudioTrackAlbumName((const uint16*)audioTrackAlbumName, audioTrackAlbumNameLength);
}


void VisualMainAction::setAudioTrackComposer(const void* const audioTrackComposer, int audioTrackComposerLength) {
	VisualDataStore::setAudioTrackComposer((const uint16*)audioTrackComposer, audioTrackComposerLength);
}


void VisualMainAction::setAudioStreamTitle(const void* const audioStreamTitle, int audioStreamTitleLength) {
	VisualDataStore::setAudioStreamTitle((const uint16*)audioStreamTitle, audioStreamTitleLength);
}


void VisualMainAction::setAudioStreamMessage(const void* const audioStreamMessage, int audioStreamMessageLength) {
	VisualDataStore::setAudioStreamMessage((const uint16*)audioStreamMessage, audioStreamMessageLength);
}


void VisualMainAction::setAudioStreamURL(const void* const audioStreamURL, int audioStreamURLLength) {
	VisualDataStore::setAudioStreamURL((const uint16*)audioStreamURL, audioStreamURLLength);
}


void VisualMainAction::setAudioStreamName(const void* const audioStreamName, int audioStreamNameLength) {
	VisualDataStore::setAudioStreamName((const uint16*)audioStreamName, audioStreamNameLength);
}


void VisualMainAction::setAudioTrackSizeInBytes(int audioTrackSizeInBytes) {
	VisualDataStore::setAudioTrackSizeInBytes(audioTrackSizeInBytes);
}


void VisualMainAction::setAudioTrackYear(int aYear) {
	VisualDataStore::setAudioTrackYear(aYear);
}


void VisualMainAction::setAudioDataIsStream(bool isStream) {
	VisualDataStore::setAudioDataIsStream(isStream);
}


bool VisualMainAction::analyzeCurrentlySetMetadata() {
	bool audioTrackDidChange = VisualDataStore::analyzeCurrentlySetMetadata();
	return audioTrackDidChange;
}


void VisualMainAction::setTotalTimeOfCurrentTrack(int totalTimeInMS) {
	VisualAudioLab* theVisualAudioLab = VisualAudioLab::getInstance();
	theVisualAudioLab->setTotalTimeOfCurrentTrack(totalTimeInMS);
}


void VisualMainAction::setStartTimeOfCurrentTrack(int startTimeInMS) {
	VisualAudioLab* theVisualAudioLab = VisualAudioLab::getInstance();
	theVisualAudioLab->setStartTimeOfCurrentTrack(startTimeInMS);
}


void VisualMainAction::setStopTimeOfCurrentTrack(int stopTimeInMS) {
	VisualAudioLab* theVisualAudioLab = VisualAudioLab::getInstance();
	theVisualAudioLab->setStopTimeOfCurrentTrack(stopTimeInMS);
}


void VisualMainAction::postTrackInfoOfCurrentAudioData() {
	const VisualString trackInfo = VisualDataStore::getInfoOfCurrentAudioDataForDisplay();
	if (trackInfo.getNumberOfNonWhitespaceCharacters() > 0) {
	
		VisualNotification trackInfoStringStyleNotification;
		trackInfoStringStyleNotification.setKey(kGetTrackInfoStringStyleMsg);
		VisualStageControl::dispatchNotification(trackInfoStringStyleNotification);
		
		VisualStyledString styledTrackInfoString;
		styledTrackInfoString.initWithVisualString(trackInfo);
		const VisualStringStyle* const trackInfoStringStyle = dynamic_cast<const VisualStringStyle* const>(trackInfoStringStyleNotification.getObject());
		if (trackInfoStringStyle) {
			styledTrackInfoString.setStyle(*trackInfoStringStyle);
		}
		
		VisualNotification styledTrackInfoStringNotification(VisualDataStore::styledTrackInfoStringId);
		styledTrackInfoStringNotification.setKey(kStyledStringMsg);
		styledTrackInfoStringNotification.setObject(styledTrackInfoString);
		styledTrackInfoStringNotification.post();
	}
}


void VisualMainAction::requestAlbumCoverArtworkOfCurrentAudioData()
{
    OSStatus aStatus = PlayerRequestCurrentTrackCoverArt(VisualSignature::getAppCookie(), VisualSignature::getAppProc());
	if (aStatus != noErr)
    {
        char errLog[512];
#if TARGET_OS_MAC
        sprintf(errLog, "Warn: Album cover artwork request failed: %s: %s (line: %d) [%s])", GetMacOSStatusErrorString(aStatus), __FILE__, __LINE__, __FUNCTION__);
#endif
#if TARGET_OS_WIN
        sprintf(errLog, "Warn: Album cover artwork request failed: %d, %s (line: %d) [%s])", aStatus, __FILE__, __LINE__, __FUNCTION__);
#endif
        writeLog(errLog);
    }
}


void VisualMainAction::postAlbumCoverArtworkOfCurrentAudioData(VISUAL_PLATFORM_DATA coverArtImageData, uint32 coverArtDataSizeInBytes, uint32 coverArtImageFormat) {
#if TARGET_OS_MAC
    const char *coverArtDataPtr = (const char*)CFDataGetBytePtr(coverArtImageData);
#endif
#if TARGET_OS_WIN
	const char* coverArtDataPtr = (const char*)coverArtImageData;
#endif
    VisualImage *aCoverArtImage = new VisualImage;
    aCoverArtImage->addData((void *)coverArtDataPtr, coverArtDataSizeInBytes);
    aCoverArtImage->dataLoadDidEnd(VisualDataStore::albumCoverArtworkImageId);
}


bool VisualMainAction::postLyricsOfCurrentAudioData() {
	bool success = false;
	success = VisualDataStore::createLyricsOfCurrentTrack();
	return success;
}


void VisualMainAction::writeLogEntry(const char* const infoStr) {
	writeLog(infoStr);
}
