/*
 * Project: VizKit
 * Version: 3.3.5
 
 * Date: 20141018
 * File: VisualTextAutoScrollView.h
 *
 */

/***************************************************************************
 
 Copyright (c) 2004-2014 Heiko Wichmann (http://www.imagomat.de/vizkit)
 
 
 This software is provided 'as-is', without any expressed or implied warranty. 
 In no event will the authors be held liable for any damages
 arising from the use of this software.
 
 Permission is granted to anyone to use this software for any purpose,
 including commercial applications, and to alter it and redistribute it
 freely, subject to the following restrictions:
 
 1. The origin of this software must not be misrepresented; 
 you must not claim that you wrote the original software. 
 If you use this software in a product, an acknowledgment 
 in the product documentation would be appreciated 
 but is not required.
 
 2. Altered source versions must be plainly marked as such, 
 and must not be misrepresented as being the original software.
 
 3. This notice may not be removed or altered from any source distribution.
 
 ***************************************************************************/

#ifndef VisualTextAutoScrollView_h
#define VisualTextAutoScrollView_h


#include "VisualTypes.h"
#include "VisualGraphicTypes.h"
#include "VisualObject.h"

#include <vector>


namespace VizKit {
    
    class VisualStyledString; // Forward declaration (to avoid include of header file).
    class VisualImage; // Forward declaration (to avoid include of header file).
    class VisualStringStyle; // Forward declaration (to avoid include of header file).
	
	/**
	 * A VisualTextAutoScrollView displays rendered text. 
     * Text that does not fit inside the view bounds is scrolled in automatically.
     * Scrolling speed and scrolling position depends on the auto play position that is passed in with the draw method.
     * Only text lines currently being displayed are rendered into images.
	 */
	class VisualTextAutoScrollView : public VisualObject {
		
	public:
		
		/**
		 * The constructor.
		 */
		VisualTextAutoScrollView();
		
		/**
		 * The destructor.
		 */
		~VisualTextAutoScrollView();
		
		/**
		 * Copy constructor.
		 * @param other Reference to another VisualTextAutoScrollView.
		 */
		VisualTextAutoScrollView(const VisualTextAutoScrollView& other);
		
		/**
		 * Assignment operator.
		 * @param other Reference to another VisualTextAutoScrollView.
		 */
		VisualTextAutoScrollView& operator=(const VisualTextAutoScrollView& other);
		
		/**
		 * Less than operator.
		 * @param other Another VisualTextAutoScrollView.
		 * @remarks The less than operator is used by std::map for find() and sorting (if VisualTextAutoScrollView is used as key_type).
		 */
		bool operator<(const VisualTextAutoScrollView& other) const;
		
		/**
		 * Equality operator.
		 * @param other Another VisualTextAutoScrollView.
		 */
		bool operator==(const VisualTextAutoScrollView& other) const;
		
		/**
		 * Inequality operator.
		 * @param other Another VisualTextAutoScrollView.
		 */
		bool operator!=(const VisualTextAutoScrollView& other) const;
		
		/**
		 * Copies the current VisualTextAutoScrollView and returns a pointer to a new VisualTextAutoScrollView.
		 */
		virtual VisualTextAutoScrollView* clone(void) const;
        
        /**
		 * Sets the visual string.
		 * @param aStyledString Reference to a styled string.
		 */
        void setStyledString(VisualStyledString& aStyledString);
        
        /**
		 * Returns the styled string.
		 * @return The styled string.
		 */
        const VisualStyledString* const getStyledString(void);
        
        /**
		 * Sets the style of the visual string.
		 * @param aStringStyle A string style.
		 */
        void setStringStyle(VisualStringStyle aStringStyle);
        
        /**
		 * Returns the current string style.
		 * @return The current string style.
		 */
        VisualStringStyle getStringStyle(void);
        
        /**
		 * Draws the visual string.
         * @param aCurrAudioPlayProgressValue Progress of the audio track (range: 0.0 to 1.0).
		 */
        void draw(float aCurrAudioPlayProgressValue);
        
        /**
		 * Refreshes the layout based on current canvas size.
		 */
        void refreshLayout(void);
        
        /**
		 * Returns the pixel rect of the view.
		 */
        PixelRect getViewRect(void);
        
        /**
		 * Sets the height of the displayed view in pixels inside which the content is clipped.
         * @param aDisplayedPixelHeight The maximum pixel height of the view content.
		 */
        void setDisplayedPixelHeight(size_t aDisplayedPixelHeight);
        
    private:
        
        class VisualTextAutoScrollViewTextLineImageRecord {
        public:
            VisualImage *textImage; /**< The image of the text line (can be empty). */
            PixelRect textImagePixelRect; /**< The width and height dimension of the image. */
            size_t textImageVerticalPixelPosition; /**< The vertical position of the text line image (sum of previous text line heights). */
        };
		
		/**
		 * Copy method for assignment operator and copy constructor.
		 * @param other Another VisualTextAutoScrollView.
		 */
		void copy(const VisualTextAutoScrollView& other);
        
        /** The styled string to be displayed in the scroll view. */
        VisualStyledString *textScrollViewStyledString;
        
        /** The vector with image data per text line. */
        std::vector<VisualTextAutoScrollViewTextLineImageRecord *> textScrollViewImageVector;
        
        /**
		 * Loads as much image data as necessary.
         * @param aProgressValue Progress of the scrolling (range: 0.0 to 1.0).
		 */
        void refreshTextlineImageData(float aProgressValue);
        
        /**
		 * Calculates the first and the last index of text lines to show.
         * @param aProgressValue Progress of the scrolling (range: 0.0 to 1.0).
		 */
        void calculateFirstAndLastTextLineIndexToShow(float aProgressValue);
        
        /**
		 * Converts the passed-in progress values of audio play progress to the internally used progress value for auto-scrolling.
         * @param aCurrAudioPlayProgressValue Progress of the audio track (range: 0.0 to 1.0).
		 */
        float convertProgressValueToInternalProgressValue(float aCurrAudioPlayProgressValue);
        
        /** The index of the first text line that needs to be shown (and needs to be loaded as image). */
        size_t firstTextLineIndexToShow;

        /** The index of the last text line that needs to be shown (and needs to be loaded as image). */
        size_t lastTextLineIndexToShow;
        
        /** The pixel width of the longest text line image. */
        size_t longestTextLinePixelWidth;
        
        /** The pixel height of all text lines summed up. */
        size_t completeTextLinesPixelHeight(void);
        
        /** The maximum number of vertically displayed pixels. */
        size_t displayedPixelHeight;
        
    };

}

#endif /* VisualTextAutoScrollView_h */
